import json
import logging
import argparse
import numpy as np
import pandas as pd

from typing import List
from pathlib import Path
from utils.other_utils import setup_logger, compute_metrics
import utils.data_utils as data_utils
from code_generator import PythonGenerator, SQLGenerator
from llm_pr.llmpr import LLMPR


logger = logging.getLogger(__name__)

def sample_demonstrations(args, task, train_data):
    # we sample based on tasks
    if task in ["entity_matching", "error_detection_spelling"]:
        demonstrations, demon_df = data_utils.sample_data_stratified(train_data, args.k)
    elif task in ["data_transformation", "data_imputation"]:
        demonstrations, demon_df = data_utils.sample_data_random(train_data, args.k) 
    else:
        raise NotImplementedError
    logger.info(demonstrations)
    return demonstrations, demon_df

def function_generation(args, train_data, instruction, task, llm: str="gpt-4-turbo", lang: str="sql"):
    saved_funcs = []
    saved_vfuncs = []
    saved_acc = []
    if lang == "python":
        generator = PythonGenerator(task=task, llm=llm, use_data_router=args.use_data_router)
        if task == "data_transformation":
            dummy_func = "def string_transformation(input_string): return None"
        elif task == "entity_matching":
            dummy_func = "def entity_matcher(input_string): return None"
        elif task == "data_imputation":
            dummy_func = "def impute_data(input_string): return None"
        elif task == "error_detection_spelling":
            dummy_func = "def detect_error(input_string): return None"
        else:
            raise NotImplementedError
    elif lang == "sql":
        generator = SQLGenerator(task=task, llm=llm, use_data_router=args.use_data_router)
        dummy_func = "SELECT 'None'"
    else:
        raise NotImplementedError
    # We do N trials 
    for trial_num in range(args.num_trials):
        seed = args.seed + trial_num
        np.random.seed(seed)
        num_try = 0
        acc = None
        t_func = "No function"
        v_func = None
        while "No function" in t_func:
            if num_try >= args.num_retry:
                # use dummy function string
                logger.info("Using dummy function.")
                t_func = dummy_func
                continue
            demonstrations, demon_df = sample_demonstrations(args, task, train_data)
            result = generator.pipeline(instruction=instruction, examples=demonstrations, depth=args.num_iter)
            t_func = result["transformation_code"]
            v_func = result["validation_code"]
            acc = result["acc"]
            num_try += 1
            seed += 1234
            np.random.seed(seed)
        logger.info("Learned function for trial {} is {}".format(trial_num, t_func))
        if args.use_data_router:
            logger.info("Validation function is {}".format(v_func))
        
        saved_funcs.append(t_func)
        saved_vfuncs.append(v_func)
        saved_acc.append(acc)
    return saved_funcs, saved_vfuncs, saved_acc, generator, demonstrations, demon_df


def main():
    # get arguments
    parser = argparse.ArgumentParser(description="Run wrangler")
    parser.add_argument(
        "--data_dir",
        type=str,
        help="Which data directory to run.",
        required=True,
    )
    parser.add_argument(
        "--output_dir", type=str, help="Output directory.", default="outputs"
    )
    parser.add_argument("--k", type=int, help="Number examples in prompt", default=3)
    parser.add_argument("--d", type=int, help="Number examples for training a classifier", default=100)
    parser.add_argument("--num_iter", type=int, help="Number of iterations to sample from training data", default=5)
    parser.add_argument("--num_retry", type=int, help="Number of retry to LLM", default=1)
    parser.add_argument(
        "--num_run",
        type=int,
        help="Number examples to run through model.",
        default=-1,
    )
    parser.add_argument(
        "--lang",
        type=str,
        help="Language used for code generation, either python or sql",
        default="sql",
    )
    parser.add_argument(
        "--llm",
        type=str,
        help="Language model used for code generation, can be openai or local",
        default="gpt-4-turbo",
    )
    parser.add_argument(
        "--num_trials",
        type=int,
        help="Number trials to run. Results will be averaged with variance reported.",
        default=1,
    )
    parser.add_argument(
        "--sample_method",
        type=str,
        help="Example generation method",
        default="random",
        choices=["random", "manual", "validation_clusters"],
    )
    parser.add_argument(
        "--class_balanced",
        help="Class balance training data. Good for classification tasks \
             with random prompts.",
        action="store_true",
    )
    parser.add_argument(
        "--use_data_router",
        help="Use data router for generating functions.",
        action="store_true",
    )
    parser.add_argument(
        "--use_fallback",
        help="Use LLMPR as fallback solution.",
        action="store_true",
    )
    parser.add_argument("--seed", type=int, default=1234)
    parser.add_argument(
        "--sep_tok",
        type=str,
        help="Separate for attr: val pairs in row. Default is '^'.",
        default="^",
    )
    parser.add_argument(
        "--nan_tok",
        type=str,
        help="Token to represent nan entries. Default is 'nan'.",
        default="nan",
    )
    args = parser.parse_args()
    # Get absolute path
    args.data_dir = str(Path(args.data_dir).resolve())
    setup_logger(args.output_dir)
    logger.info(json.dumps(vars(args), indent=4))
    limit_fallback = 100
    
    llm = args.llm
    logger.info(f"Using LLM: {llm}")

    test_file = "test"

    # Read pandas DF datasets
    pd_data_files, task = data_utils.read_data(
        data_dir=args.data_dir,
        class_balanced=args.class_balanced,
        add_instruction=False,
        max_train_samples=-1,
        max_train_percent=-1,
        sep_tok=args.sep_tok,
        nan_tok=args.nan_tok,
    )
    logger.info(f"Task for the dataset is {task}")
    if test_file not in pd_data_files:
        raise ValueError(f"Need {test_file} data")

    train_data = pd_data_files["train"]
    test_data = pd_data_files[test_file]
    num_run = args.num_run
    instructions = []

    if isinstance(train_data, List):
        if args.num_run == -1:
            num_run = test_data[0].shape[0]
        num_run = min(num_run, test_data[0].shape[0])
        logger.info(f"Number of tasks is {len(train_data)}")
        logger.info(f"Train shape is {train_data[0].shape[0]}")
        logger.info(f"Test shape is {test_data[0].shape[0]}")
        logger.info(f"Running {num_run} examples for {args.num_trials} trials.")
        if "instructions" in pd_data_files:
            instructions = pd_data_files["instructions"]
        logger.info(f"Number of tasks is {len(instructions)}")

    elif isinstance(train_data, pd.DataFrame):
        if args.num_run == -1:
            num_run = test_data.shape[0]
        num_run = min(num_run, test_data.shape[0])
        logger.info(f"Train shape is {train_data.shape[0]}")
        logger.info(f"Test shape is {test_data.shape[0]}")
        logger.info(f"Running {num_run} examples for {args.num_trials} trials.")
        train_data = [train_data]
        test_data = [test_data]
        instruction = data_utils.read_instruction(args.data_dir)
        if instruction:
            instructions = [instruction]
        else:
            instructions = [""]

    task_metrics = {"prec": [], "rec": [], "f1": [], "acc": []}
    all_gts = []
    all_preds = []
    all_saved_funcs = []
    task_number = 0
    for train_data_pd, test_data_pd, instruction in zip(train_data, test_data, instructions):
        task_number += 1
        logger.info(f"Task instruction {instruction}")
        if len(test_data_pd) == 0:
            logger.info("Not enough samples to run, continue to next task.")
            continue
        saved_funcs = []
        # remove previous demonstrations records from train_data_pd

        
        # we have a list of dataframes where each dataframe represent a task 
        saved_funcs, saved_vfuncs, saved_acc, generator, demonstrations, demon_df = function_generation(args, train_data_pd, instruction, task=task, llm=llm, lang=args.lang)
        print(demonstrations)
        all_saved_funcs.append(saved_funcs)
        batches = []
        test_data_lst = data_utils.deserialize_data(test_data_pd)
        batches = [test_data_lst]
        
        accs_per_batch = []
        gts_per_task = []
        preds_per_task = []
        invalid_data_per_batch = []
        i = 0
        logger.info(f"number of batches splitted is {len(batches)}")
        logger.info(f"number of funcs is {len(saved_funcs)}")

        # evaluate per batch
        if task in ["data_transformation", "error_detection_spelling"]:
            num_solution = None
            for batch, func, vfunc in zip(batches, saved_funcs, saved_vfuncs):
                logger.info(f"using function {func} to evaluate")
                
                # when there is no validation function, we route all data to the current function
                valid_data, invalid_data = generator.route_data(vfunc, batch)
                invalid_data_per_batch.append(invalid_data)
                acc, preds = generator.evaluate(func, valid_data)
               
                logger.info(f"acc: {acc} for batch {i}")
                accs_per_batch.append(acc)
                for sample, pred in zip(valid_data, preds):
                    gt = sample["Output"]
                    gts_per_task.append(gt)
                    preds_per_task.append(pred)
                    all_gts.append(gt)
                    all_preds.append(pred)
                    logger.info(f"====> pred: {pred} <====")
                    logger.info(f"====> gt: {gt} <====")

                if invalid_data:
                    for invalid_data_sample in invalid_data:
                        logger.info(f"====> invalid data: {invalid_data_sample} <====")
                    # run the fallback solution
                    if args.use_fallback:
                        logger.info("Using fallback solution.")
                        llmpr = LLMPR(model_name="gpt-4-turbo", task=task)
                        gts, preds = llmpr.pipeline(examples=demonstrations, test_data=invalid_data)
                        for gt, pred in zip(gts, preds):
                            gts_per_task.append(gt)
                            preds_per_task.append(pred)
                            all_gts.append(gt)
                            all_preds.append(pred)
                            logger.info(f"====> pred: {pred} <====")
                            logger.info(f"====> gt: {gt} <====")                   

        else:
            # for tasks entity matching and data imputation
            sorted_acc, sorted_funcs, sorted_vfuncs = zip(*sorted(zip(saved_acc, saved_funcs, saved_vfuncs), key=lambda x: x[0], reverse=True))
            func = sorted_funcs[0]
            vfunc = sorted_vfuncs[0]
            logger.info(f"sorted acc {sorted_acc}, sorted funcs {sorted_funcs}, sorted vfuncs {sorted_vfuncs}")
            logger.info(f"Acc for the fincal func is {sorted_acc[0]}, Final function to use is {func}, Final validation function to use is {vfunc}")           
            for batch in batches:
                valid_data, invalid_data = generator.route_data(vfunc, batch)
                invalid_data_per_batch.append(invalid_data)
                acc, preds = generator.evaluate(func, valid_data)
               
                logger.info(f"acc: {acc} for batch {i}")
                
                accs_per_batch.append(acc)
                for sample, pred in zip(valid_data, preds):
                    gt = sample["Output"]
                    gts_per_task.append(gt)
                    preds_per_task.append(pred)
                    all_gts.append(gt)
                    all_preds.append(pred)
                    logger.info(f"====> pred: {pred} <====")
                    logger.info(f"====> gt: {gt} <====")
                
                # single solution strategy
                # if invalid_data:
                #     for invalid_data_sample in invalid_data:
                #         logger.info(f"====> invalid data: {invalid_data_sample} <====")
                #     if not valid_data:
                #         logger.info("All data is invalid. We don't run the fallback solution.")
                #     else:
                #         # run the fallback solution
                #         # set up maximum number of invalid data to run the fallback solution 
                #         if args.use_fallback & len(invalid_data) < 100:
                #             logger.info("Using fallback solution.")
                #             llmpr = LLMPR(model_name="gpt-4-turbo", task=task)
                #             gts, preds = llmpr.pipeline(examples=demonstrations, test_data=invalid_data)
                #             for gt, pred in zip(gts, preds):
                #                 gts_per_task.append(gt)
                #                 preds_per_task.append(pred)
                #                 all_gts.append(gt)
                #                 all_preds.append(pred)
                #                 logger.info(f"====> pred: {pred} <====")
                #                 logger.info(f"====> gt: {gt} <====")  
                # 
                # Multiple solution strategy
                # the number of solutions should be determined by the length of invalid data in the future. 
                max_num_solutions = 5
                num_solution = 1
                while len(invalid_data) > limit_fallback:
                    logger.info(f"===========>Current number of solutions is {num_solution}<===========")
                    if num_solution >= max_num_solutions:
                        break
                    logger.info(f"length of invalid data is {len(invalid_data)}, generate new functions")
                    saved_funcs = []

                    # before regenerate another code solution, remove the previous demonstrations records from train_data_pd
                    train_data_pd = train_data_pd.drop(index=demon_df.index)
                    logger.info(f"length of train data after removing demonstrations is {train_data_pd.shape[0]}")
                    # we have a list of dataframes where each dataframe represent a task 
                    saved_funcs, saved_vfuncs, saved_acc, generator, demonstrations, demon_df = function_generation(args, train_data_pd, instruction, task=task, llm=llm, lang=args.lang)
                    sorted_acc, sorted_funcs, sorted_vfuncs = zip(*sorted(zip(saved_acc, saved_funcs, saved_vfuncs), key=lambda x: x[0], reverse=True))
                    func = sorted_funcs[0]
                    vfunc = sorted_vfuncs[0]
                    logger.info(f"sorted acc {sorted_acc}, sorted funcs {sorted_funcs}, sorted vfuncs {sorted_vfuncs}")
                    logger.info(f"Acc for the fincal func is {sorted_acc[0]}, Final function to use is {func}, Final validation function to use is {vfunc}")        
                    valid_data, invalid_data = generator.route_data(vfunc, invalid_data)
                    invalid_data_per_batch.append(invalid_data)
                    acc, preds = generator.evaluate(func, valid_data)
                    
                    logger.info(f"acc: {acc} for batch {i}")
                    
                    accs_per_batch.append(acc)
                    for sample, pred in zip(valid_data, preds):
                        gt = sample["Output"]
                        gts_per_task.append(gt)
                        preds_per_task.append(pred)
                        all_gts.append(gt)
                        all_preds.append(pred)
                        logger.info(f"====> pred: {pred} <====")
                        logger.info(f"====> gt: {gt} <====")
                    num_solution += 1



                if len(invalid_data) < limit_fallback:
                    logger.info(f"length of invalid data is {len(invalid_data)}, run the fallback solution")
                    logger.info("Using fallback solution.")
                    llmpr = LLMPR(model_name="gpt-4-turbo", task=task)
                    gts, preds = llmpr.pipeline(examples=demonstrations, test_data=invalid_data)
                    for gt, pred in zip(gts, preds):
                        gts_per_task.append(gt)
                        preds_per_task.append(pred)
                        all_gts.append(gt)
                        all_preds.append(pred)
                        logger.info(f"====> pred: {pred} <====")
                        logger.info(f"====> gt: {gt} <====")  
                
                else:
                    logger.info(f"Remaining invalid data size is {len(invalid_data)}. We don't run the fallback solution.")
                    for invalid_data_sample in invalid_data:
                        logger.info(f"====> invalid data: {invalid_data_sample} <====")

        logger.info(f"Number of solutions generated before running the fallback solution is {num_solution}")
        # calculating metrics per task
        prec, rec, acc, f1 = compute_metrics(preds_per_task, gts_per_task, task)
        logger.info(f"Task number {task_number}")
        logger.info(
            f"Prec: {prec:.3f} Recall: {rec:.3f} Acc: {acc:.3f} F1: {f1:.3f}"
        )
    
        task_metrics["rec"].append(rec)
        task_metrics["prec"].append(prec)
        task_metrics["acc"].append(acc)
        task_metrics["f1"].append(f1)


    output_file = (
        Path(args.output_dir)
        / f"{Path(args.data_dir).stem}"
        / f"{test_file}"
        / f"{args.k}k"
        / f"{args.d}d"
        f"_data_router_{int(args.use_data_router)}"
        f"_fallback_{int(args.use_fallback)}"
        f"_lang_{args.lang}"
        f"_llm_{llm}"
        f"_{int(args.class_balanced)}cb"
        f"_{args.sample_method}"
        f"_{args.num_run}run" / f"trial_{num_run}.feather"
    )

    output_file.parent.mkdir(parents=True, exist_ok=True)
    logger.info(f"Saved to {output_file}")
    
    # calculate metrics all rows
    prec, rec, acc, f1 = compute_metrics(all_preds, all_gts, task=task)
    task_metrics["prec_on_rows"] = prec
    task_metrics["rec_on_rows"] = rec
    task_metrics["acc_on_rows"] = acc
    task_metrics["f1_on_rows"] = f1

    for k, values in list(task_metrics.items()):
        task_metrics[f"{k}_avg"] = np.average(values)
        task_metrics[f"{k}_std"] = np.std(values)

    output_metrics = output_file.parent / "metrics.json"
    json.dump(task_metrics, open(output_metrics, "w"))

    output_functions = output_file.parent / "learned_funcs.json"
    json.dump(all_saved_funcs, open(output_functions, "w"))
    logger.info(f"Final Metrics {json.dumps(task_metrics, indent=4)}")
    logger.info(f"Metrics dumped to {output_metrics}")
    logger.info(f"Learned funcs dumped to {output_functions}")



if __name__ == "__main__":
    main()


import ast
import sys
sys.path.insert(0, '')
import duckdb
import logging
import multiprocessing
import traceback
from pydantic import BaseModel

from .base_generator import CodeGenerator
from src.tableswift.utils import function_utils
from src.tableswift.utils.formatter_utils import format_model_output_sql

logger = logging.getLogger(__name__)

class SQLResponse(BaseModel):
    reasoning: str
    transformation_code: str

class SQLResponse_DataRouter(BaseModel):
    reasoning: str
    transformation_code: str
    validation_code: str

# Move the query_runner function to the top level
def query_runner(fn, input_string, fn_name, result_dict):
    """
    Executes the DuckDB query in a separate process.
    """
    try:
        # Execute the macro function
        duckdb.sql(fn)
        # Execute the query
        query = f"SELECT {fn_name}('{input_string}')"
        result = duckdb.sql(query).fetchall()[0][0]
        result_dict['result'] = result
    except Exception as e:
        result_dict['error'] = traceback.format_exc()

class TimeoutException(Exception):
    pass

class SQLGenerator(CodeGenerator):
    def __init__(self, task: str, llm: str, use_data_router: bool, api_key: str) -> None:
        super().__init__(task, llm, api_key)
        self.lang = "sql"
        self.use_data_router = use_data_router
    
    def get_response(self, messages):
        """
        We use instructor and pre-defined response model
        """
        print("here is the complete messages, ", messages)
        response_model = SQLResponse if not self.use_data_router else SQLResponse_DataRouter
        structured_output = self.call_llm(messages=messages, response_model=response_model)
        if self.llm in self.ollama_models or self.llm in self.together_models:
            structured_output = format_model_output_sql(structured_output)
            print("Structured output after reformatting is ", structured_output)
        return structured_output
    
    def pre_process(self, input: str):
        if isinstance(input, str):
            input = input.replace("'", "''")
        return input
    
    def post_process(self, structured_output):
        """
        Post process the structured output, return a dictionary
        """
        result = {"resoning": None, "transformation_code": None, "validation_code": None}
        if self.llm in self.ollama_models or self.llm in self.together_models:
            result["transformation_code"] = structured_output
            return result
        if not self.use_data_router:
            try:
                reasoning = structured_output.reasoning
                result["reasoning"] = reasoning
                transformation_code = structured_output.transformation_code
                if transformation_code:
                    transformation_code = transformation_code.replace('\\\\', '\\')
                    result["transformation_code"] = transformation_code
            except:
                pass
            return result
        else:
            try:
                reasoning = structured_output.reasoning
                result["reasoning"] = reasoning
                transformation_code = structured_output.transformation_code
                validation_code = structured_output.validation_code
                if transformation_code:
                    transformation_code = transformation_code.replace('\\\\', '\\')
                    result["transformation_code"] = transformation_code
                if validation_code:
                    validation_code = validation_code.replace('\\\\', '\\')
                    result["validation_code"] = validation_code
            except:
                pass
            return result

    def execute_function_string(self, fn, input_string, fn_type="transformation", timeout=5):
        # TODO: parse to get the function name.
        # Prepare a custom namespace for executing the dynamic code. This can be an empty dict.
        if fn_type == "transformation":
            fn_name = {
                "data_transformation": "string_transformation",
                "entity_matching": "input_output_converter",
                "data_imputation": "input_output_converter",
                "error_detection_spelling": "detect_error",
            }.get(self.task, None)
        elif fn_type == "validation":
            fn_name = "validate"
        if not fn_name:
            raise NotImplementedError(f"Task '{self.task}' not supported.")
        
        preprocessed_input = self.pre_process(input_string)
        # Use a multiprocessing manager to share data between processes
        with multiprocessing.Manager() as manager:
            result_dict = manager.dict()

            # Create a new process to run the query
            process = multiprocessing.Process(
                target=query_runner,
                args=(fn, preprocessed_input, fn_name, result_dict)
            )
            process.start()

            # Wait for the process to finish or timeout
            process.join(timeout)

            if process.is_alive():
                # Terminate the process if it's still running
                process.terminate()
                process.join()
                raise TimeoutException("Query execution exceeded timeout and was terminated.")
            # Check for errors
            if 'error' in result_dict:
                raise Exception(result_dict['error'])
            return result_dict.get('result')

    def execute(self, fn, test_data, fn_type="transformation", timeout=5):
        """
        Wrapper to execute the SQL function with error handling.
        """
        try:
            result = self.execute_function_string(fn, test_data[0]['Input'], fn_type, timeout)
            return True, f"Execution successful. Result: {result}"
        except TimeoutException as te:
            return False, str(te)
        except Exception as e:
            return False, f"Error: {str(e)}"
    
    def evaluate(self, fn, test_data):
        ### naive evaluation without comparing 
        inputs = [item["Input"] for item in test_data]
        outputs = [item["Output"] for item in test_data]
        predicted_outputs = []
        for input, output in zip(inputs, outputs):
            try:
                result = self.execute_function_string(fn, input)
            except Exception as error:
                print(f"Error message: {error}")
                result = "Not excutable"
            predicted_outputs.append(result)
            print("pred: {}, gt: {}".format(result, output))
        # acc = calculate_accuracy(predicted=predicted_outputs, ground_truth=outputs)
        metrics = function_utils.compute_metrics(preds=predicted_outputs, golds=outputs, task=self.task)
        if self.task in [
                "data_imputation",
                "data_transformation",
            ]:
            acc = metrics[2]
        elif self.task in ["entity_matching", "error_detection_spelling"]:
            acc = metrics[3]
        print("The accuacry/f1 of the generated function is ", acc)
        return acc, predicted_outputs
    
    def route_data(self, v_fn, test_data):
        """
        Route the data given the response from the validation function
        """
        valid_data = []
        invalid_data = []
        if v_fn:
            inputs = [item["Input"] for item in test_data]
            outputs = [item["Output"] for item in test_data]
            for input, output in zip(inputs, outputs):
                try:
                    decision = self.execute_function_string(v_fn, input, "validation")
                    logger.info(f"======> Decision for input {input} is {decision} <======")
                    if decision:
                        valid_data.append({"Input": input, "Output": output})
                    else:
                        invalid_data.append({"Input": input, "Output": output})
                except Exception as error:
                    # if the validation function is not excutable, we treat it as valid data
                    print(f"Error message: {error}")
                    valid_data.append({"Input": input, "Output": output})
        else:
            valid_data = test_data
        return valid_data, invalid_data
    
    def pipeline_backbone(self, messages, examples, supervision_data, depth):
        """
        Pipeline for getting a valid transformation function
        """
        func = None
        num_iter = 1
        functions_stack = []
        while num_iter < depth:
            # getting the first function
            response = self.get_response(messages)
            result = self.post_process(response)
            if result["transformation_code"]:
                func = result["transformation_code"]
                decision, msg, acc = self.validate(func, examples, supervision_data)
                # for debugging
                print("The decision is ", decision)
                print("The message is ", msg)
                print("The accuracy is ", acc)

                if "not" in decision:
                    if not decision == "not_excutable":
                        functions_stack.append((func, acc))
                    # if this is the last try and the function runs then we return
                    if (num_iter == depth - 1):
                        if functions_stack:
                            # sort the function stack
                            sorted_func_stack = sorted(functions_stack, key = lambda x: x[1], reverse=True)
                            return (*sorted_func_stack[0], result)
                    # add error message and re-prompt
                    retry_msg = self.prompts_set["TASK_RETRY"][0].copy()
                    retry_msg['content'] = retry_msg['content'].format(error_message=msg)
                    messages.append(retry_msg)     
                else:
                    return (func, acc, result)           
            else:
                decision = "not parsable"
                msg = "The function calling return not parsable response, trying again!"
                acc = 0
                 # add error message and re-prompt
                retry_msg = self.prompts_set["TASK_RETRY"][0].copy()
                retry_msg['content'] = retry_msg['content'].format(error_message=msg)
                messages.append(retry_msg)   
            num_iter += 1
        return ("No function can be generated, please provide different demonstration!", float("-inf"), result)

    def pipeline_data_router(self, messages, examples, supervision_data, depth):
        """
        Pipeline with data router
        """
        # first prioritize the transformation code, after getting a valid transformation code, we then validate if the validation code is valid
        # for now we only valid the validation function but not re-generate
        # TODO: in the future we can re-generate the validation function
        t_func = None
        t_func, acc, result = self.pipeline_backbone(messages, examples, supervision_data, depth)
        if not t_func:
            return t_func, acc, None
        # now we validate the validation function
        v_func = result["validation_code"]
        if not v_func:
            return t_func, acc, None
        runnable, msg = self.execute(v_func, examples, "validation")
        if runnable:
            return t_func, acc, v_func
        else:
            return t_func, acc, None

    def pipeline(self, instruction, examples, supervision_data=None, depth: int=5):
        """
        The sequence of action for generating the final function is: generate function -> validate function -> correct function if necessary.
        """
        result = {"transformation_code": None, "acc": None, "validation_code": None}
        # Define the messages
        messages, _ = self.formulate_prompt(instruction=instruction, examples=function_utils.dicts_to_string(examples))   
        if not self.use_data_router:
            code, acc, _ = self.pipeline_backbone(messages, examples, supervision_data, depth)
            result["transformation_code"] = code
            result["acc"] = acc
        else:
            # use pipeline that contains data router
            code, acc, v_code = self.pipeline_data_router(messages, examples, supervision_data, depth)
            result["transformation_code"] = code
            result["acc"] = acc
            result["validation_code"] = v_code
        return result


# do we really want to do execution in the ts framework?
from ..code_generator import PythonGenerator, SQLGenerator
from ..llm_pr.llmpr import LLMPR

def execute_code_logic(code: str, lang: str, inputs: list,
                       instruction: str, 
                       demonstrations: list, task: str,
                       hyperparams: dict, api_key: str, router_code: str = None) -> None:
    """
    Execute the generated code.
    """
    print("Executing code logic...")
    print(f"Executing Code: {code}")
    print(f"Language: {lang}")
    print(f"Router Code: {router_code}")

    llm = hyperparams["llm"]
    use_data_router = hyperparams["use_data_router"]
    max_num_solutions = hyperparams["max_num_solutions"]
    limit_fallback = hyperparams["limit_fallback"]

    transformation_data = []
    if lang == "python":
        generator = PythonGenerator(task=task, llm=llm, use_data_router=use_data_router, api_key=api_key)
    elif lang == "sql":
        generator = SQLGenerator(task=task, llm=llm, use_data_router=use_data_router, api_key=api_key)
    else:
        raise NotImplementedError(f"Language {lang} is not supported for execution.")
    # inputs should be a list of dictionaries that shows {"Input": "sample1", "Output": ""}
    valid_data, invalid_data = generator.route_data(router_code, inputs)
    # invalid_data_per_batch.append(invalid_data)
    for sample in valid_data:
        sample_input = sample["Input"]
        result = generator.execute_function_string(code, sample_input)
        sample["Output"] = result
        transformation_data.append(sample)
    # valid_data should be a list of dictionaries that shows {"Input": "sample1", "Output": "label1"}
    # there is still logic on how to handle invalid data and defining the maximum number of solutions
    if len(invalid_data) == 0:
        print("All samples were successfully processed.")
    elif len(invalid_data) <= limit_fallback:
        print(f"Fallback triggered: {len(invalid_data)} invalid samples left, within the limit of {limit_fallback}.")
        # fall back to LLMPR
        llm_pr = LLMPR(task=task, api_key=api_key, model_name=llm)
        transformed_samples = llm_pr.pipeline_no_eval(instruction, demonstrations, invalid_data)
        transformation_data.extend(transformed_samples)
    else:
        print(f"Too many invalid samples left: {len(invalid_data)}, should generate another code.")
    return transformation_data, invalid_data


# here we implement the fallback solution, using LLM on a per-row basis.
import copy
import openai
import logging
import instructor

from pydantic import BaseModel
from src.tableswift.utils import function_utils


logger = logging.getLogger(__name__)

class LLMPRResponse(BaseModel):
    output: str

prefix_instr_dt = [{"role": "system", "content": "You are helpful assistant. Given the following input/output examples, figure out the transformation and generate the output when given input."}]
prefix_instr_em = [{"role": "system", "content": "You are helpful assistant. Given the following entities, figure out if entity A is the same with entity B. Yes or No?"}]
demonstration_instr = [{"role": "user", "content": f"""Instructions: {{instruction}}, Examples: {{examples}}"""}]


class LLMPR:
    def __init__(self, model_name, task, api_key) -> None:
        self.model_name = model_name
        self.task = task
        self.api_key = api_key

    def formulate_prompt(self, instruction, examples):
        # print(f"Current demonstration template is ", demonstration_instr)
        prefix = []
        if self.task == "data transformation":
            prefix = copy.deepcopy(prefix_instr_dt)
        elif self.task == "entity matching":
            prefix = copy.deepcopy(prefix_instr_em)
        demonstration = copy.deepcopy(demonstration_instr)
        messages = prefix + demonstration
        for message in messages:
            if 'user' in message['role']:
                message['content'] = message['content'].format(instruction=instruction, examples=examples) 
        return messages

    def call_llm(self, message_row):
        # we run the LLM on the invalid rows
        client = instructor.from_openai(client = openai.OpenAI(api_key=self.api_key))
        # Extract structured data from natural language
        structured_output = client.chat.completions.create(
            model=self.model_name,
            response_model=LLMPRResponse,
            messages=message_row,
        )
        return structured_output
    
    def execute(self, messages, test_data):
        """
        We run the LLM on the invalid rows, and return results. No gts.
        """
        transformed_data = []
        for row in test_data:
            message_row = copy.deepcopy(messages) + [{"role": "user", "content": f"Input: {row['Input']}\nOutput:"}]
            structured_output = self.call_llm(message_row)
            if structured_output.output:
                output_pred = structured_output.output.lower().strip()
            else:
                output_pred = structured_output.output
            row['Output'] = output_pred
            transformed_data.append(row)
        return transformed_data
        
    def evaluate(self, messages, test_data):
        all_outputs_preds = []
        all_outputs_gt = []
        logger.info(f"prompts with demonstrations {messages}")
        for row in test_data:
            message_row = copy.deepcopy(messages) + [{"role": "user", "content": f"Input: {row['Input']}\nOutput:"}]
            structured_output = self.call_llm(message_row)
            if row['Output']:
                output_row = row['Output'].lower().strip()
            else:
                output_row = row['Output']
            if structured_output.output:
                output_pred = structured_output.output.lower().strip()
            else:
                output_pred = structured_output.output
            logger.info(f"====> Input: {row['Input']}, gt: {output_row}, pred: {output_pred} <===")
            all_outputs_preds.append(output_pred)
            all_outputs_gt.append(output_row)
        return all_outputs_gt, all_outputs_preds

    def pipeline(self, instructions, examples, test_data):
        """
        We run the LLM on the invalid rows, and then calculate the accruacy.
        """
        logger.info(f"Running fallback solution on {len(test_data)} invalid rows")
        examples=function_utils.dicts_to_string(examples)
        messages = self.formulate_prompt(instructions, examples)
        gts, preds = self.evaluate(messages, test_data)
        return gts, preds
    
    def pipeline_no_eval(self, instruction, examples, test_data):
        """
        Inference piopeline without evaluation.
        """
        logger.info(f"Running fallback solution on {len(test_data)} invalid rows")
        if examples:
            examples=function_utils.dicts_to_string(examples)
        else:
            examples = ""
        messages = self.formulate_prompt(instruction, examples)
        transformed_data = self.execute(messages, test_data)
        return transformed_data



        
        


import tableswift as ts
from tableswift import generate_labels
from tableswift import generate_code



if __name__ == "__main__":
    ts.configure(api_key="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")
    
    # Generate labels for the samples
    labeled_data = ts.generate_labels(instruction="label the input samples", 
                                      task="data_transformation",
                                      demonstrations=[{"Input": "sample1", "Output": "label1"},
                                                     {"Input": "sample2", "Output": "label2"}],
                                      samples_to_label=[{"Input": "sample1", "Output": ""},
                                                        {"Input": "sample2", "Output": ""},
                                                        {"Input": "sample3", "Output": ""}])
    print("labeled_data:")
    for data in labeled_data:
        print(data)
    
    # Generate code for the package
    code, router_code = ts.generate_code(instruction="Transform input into output",
                     task="data_transformation",
                     samples=[{"Input": "sample1", "Output": "label1"},
                              {"Input": "sample2", "Output": "label2"}],
                     lang="python",
                     num_trials=1,
                     num_retry=3,
                     num_iterations=1)
    print("Generated Code:")
    print(code)
    print("Router Code:")
    print(router_code)
    # Execute the generated code
    inputs = [{"Input": "sample1", "Output": ""},
              {"Input": "sample2", "Output": ""},
              {"Input": "sample3", "Output": ""}]
    results, invalid_data = ts.execute_code(code=code,
                                            instruction="Transform input into output",
                                            task="data_transformation",
                                            lang="python",
                                            inputs=inputs,
                                            samples=[{"Input": "sample1", "Output": "label1"},
                                                     {"Input": "sample2", "Output": "label2"}],
                                            router_code=router_code)
    print("Execution Results:")
    for result in results:
        print(result)
    
    


# config file for setting up API keys for Tableswift

import os

_api_key = None
_params = {}
# Default parameters for the Tableswift API, can be overridden by user
DEFAULT_PARAMS = {
    "use_data_router": True,
    "num_trials": 2,
    "num_retry": 3,
    "seed": 42,
    "num_iterations": 2,
    "max_num_solutions": 3,
    "limit_fallback": 20, # number of invalid data samples before fallback, should be a percentage in the future
    "llm": "gpt-4o-mini" 
}

def configure(api_key: str):
    global _api_key
    _api_key = api_key

def get_api_key():
    return _api_key or os.getenv("TABLESWIFT_API_KEY")

def require_api_key():
    key = get_api_key()
    if not key:
        raise ValueError("No API key provided. Use configure() or set TABLESWIFT_API_KEY env var.")
    return key

def with_api_key(func):
    def wrapper(*args, **kwargs):
        key = require_api_key()
        return func(*args, api_key=key, **kwargs)
    return wrapper

def get_param(name):
    return _params.get(name, DEFAULT_PARAMS.get(name))

def resolve_hyperparams(overrides: dict = None):
    overrides = overrides or {}
    params = {}

    for param_name in DEFAULT_PARAMS:
        params[param_name] = overrides.get(
            param_name, 
            get_param(param_name)
        )

    return params


from typing import List
from .config import with_api_key, resolve_hyperparams
from .logic.codegen import generate_code_logic
from .logic.execution import execute_code_logic
from .logic.labeling import label_data_logic


@with_api_key
def generate_labels(instruction: str, task: str, samples_to_label: List, api_key=None, demonstrations: List[dict]=None, **overrides) -> None:
    """
    Generate labels given instructions and input samples.
    """
    # input should be a list of dictionaries
    hyperparams = resolve_hyperparams(overrides)
    return label_data_logic(
        instruction=instruction,
        task=task,
        demonstrations=demonstrations,
        samples_to_label=samples_to_label,
        api_key=api_key, 
        hyperparams=hyperparams)

@with_api_key
def generate_code(instruction: str, task:str, samples: List, lang: str, api_key=None, **overrides) -> str:
    """
    Generate code given instructions and input samples.
    """
    hyperparams = resolve_hyperparams(overrides)
    return generate_code_logic(
        lang=lang,
        task=task,
        instruction=instruction,
        samples=samples,
        api_key=api_key,
        hyperparams=hyperparams
    )

@with_api_key
def execute_code(code: str, task: str, lang: str, instruction: str, inputs: List, samples: List,
                 router_code: str=None, api_key: str=None, **overrides) -> None:
    """
    Execute the generated code.
    """
    hyperparams = resolve_hyperparams(overrides)
    return execute_code_logic(
        code=code,
        lang=lang,
        router_code=router_code,
        instruction=instruction,
        inputs=inputs,
        demonstrations=samples,
        task=task,
        hyperparams=hyperparams,
        api_key=api_key)




from typing import *

PYTHON_PROMPTS = {
    "TASK_PREFIX":[ {"role": "system", "content": "You are helpful assistant. Given the following input/output examples, generate a python function that converts input to output. Reason step-by-step about the what the function is supposed to do and then generate the function. You can assume that the following python libraries are available in the target environment."}],
    "TASK_RETRY":[ {"role": "user", "content": f"{{error_message}}"}],
    "STRING_TRANSFORMATION": [{"role": "user", "content": """You are an expert Python user. I will give you a string transformation task. \
                         The task involves converting input strings to output strings. \
                         Assuming all of the following python packages are installed: ["regex", "fraction", "math", "pyproj", "BeautifulSoup", "geopy", "ummalqura", "mgrs", "pytz", "datetime", "calendar", "roman"],\
                          you can **import** them and please put all imports in the begining of the generated python program. You can also import other packages if the listed ones are not working, but try using listed packages first. \
                          When it comes to currency exchange, please do not use real-time currency, but figure out the exchange rate based on the examples I provide.
                         You will have to find the Python program called "string_transformation" that will generate the output string given the input string. \
                          Output can only be string and nothing else. \
                          Please try to understand the intentions of the input-output examples while generating the function. \
                          Please return only the python function and nothing else.  Include any imports. Do not provide any comments.\
                          Here is a very simple example: \n
                          Input: "john.doe@example.com" \n 
                          Output: "Yes" \n 
                          Input: "jane_doe@example.com" \n 
                          Output: "Yes" 
                          Input: "example.com" \n
                          Output: "No" \n
                          Generated function string: \n
                          ```pythonimport re\ndef string_transformation(input_string):\n\tpattern = r'^[\w\.-]+@[\w\.-]+\.\w+$'\n\tif re.match(pattern, input_string):\n\t\treturn "Yes"\n\telse:
        return\n\t\t"No"```\n
                          Now it's your turn. After generating the function string, please add escape symbol tab '\t' to format the code properly with indents like shown in the example. """},
                        {"role": "user", "content": f"""Instructions: {{instruction:}}, Examples: {{examples:}}"""}],

    "ENTITY_MATCHING": [{"role": "user", "content": """I have an entity matching task to solve. I have pairs of rows and I want you to find a python program that checks if they match with each other. 
                    Please try to match using substring and regex preferably. 
                    First consider using Jaccard, levenshtein, cosine similarity or any other edit distance function to solve the problem. 
                    Try to define a low threshold so that the generated function is **generalizable**.
                    Note that there can be missing values in some columns, and you can skip them for matching. All columns are connected by the symbol ^ in the input string.
                    Please do not train any machine learning classifier.
                    You can try to choose appropriate distance function with revevant columns based on the column meaning.
                    \n
                    Assuming all of the following python packages are installed in the target environment: ["numpy", "regex", "fractions", "sklearn", "Levenshtein", "math", "pyproj", "BeautifulSoup", "geopy", "ummalqura", "mgrs", "pytz", "datetime", "calendar", "roman"], 
                    you can import them when it is necessary and please put all imports in the beginning of the generated python program. 
                    You can also import other packages if the listed ones are not working, but try using listed packages first. 
                    If you need to list some mappings, try to retrieve all entries first or write a function to retrieve it first then write the matching function. 
                    Try to generate functions that can generalize to many other data samples than the ones we provided here. 
                    \n
                    The main python function that's callable to solve the task is called "input_output_converter(input_string)" where "input_string" is the function parameter. 
                    One example function string is: ```pythonimport re\ndef input_output_converter(input_string):\n\tpattern = r'^[\w\.-]+@[\w\.-]+\.\w+$'\n\tif re.match(pattern, input_string):\n\t\treturn "Yes"\n\telse: return\n\t\t"No"``` 
                    \n
                    Please return only the python function and nothing else. Include any imports. Please do **not** provide any comments. 
                    After generating the function string, please add escape symbol tab '\t' to format the code properly with indents like shown in the example."""},\
                        {"role": "user", "content": f"""Instructions: {{instruction:}}, Examples: {{examples:}}"""}],
    "DATA_IMPUTATION": [{"role": "user", "content": """You are a helpful code searcher. I have a data imputation task to solve. I have rows of records and I want you to find a python program that substituting missing values given the column name.

                    Please try to find programs that uses logical conditions, substring and regex preferably, in combination of the column meanings. 
                    Try to define the range dynamically so that the generated program is **generalizable**. 
                    Please do not train any machine learning classifier.

                    Assuming all of the following python packages are installed in the target environment: ["regex", "fractions", "sklearn", "Levenshtein", "math", "pyproj", "BeautifulSoup", "geopy", "ummalqura", "mgrs", "pytz", "datetime", "calendar", "roman"], you can import them when it is necessary and please put all imports in the beginning of the generated python program. 
                    You can also import other packages if the listed ones are not working, but try using listed packages first. 
                    If you need to list some mappings, try to retrieve all entries first or write a function to retrieve it first then write the matching function. 
                    Try to generate functions that can generalize to many other data samples than the ones we provided here.
                    The main python function that's callable to solve the task is called "input_output_converter(input_string)" where "input_string" is the function parameter. 
                    One example function string is: pythonimport re\ndef input_output_converter(input_string):\n\tpattern = r'^[\w\.-]+@[\w\.-]+\.\w+$'\n\tif re.match(pattern, input_string):\n\t\treturn "Yes"\n\telse: return\n\t\t"No"

                    Please return only the python function and nothing else. Include any imports. Please do not provide any comments. After generating the function string, please add escape symbol tab '\t' to format the code properly with indents like shown in the example.

                    """},
                    {"role": "user", "content": f"""Instructions: {{instruction:}}, Examples: {{examples:}}"""}],
    "ERROR_DETECTION_SPELLING": [{"role": "user", "content": """You are a helpful code generator. I have an error detection task to solve. I have rows of records from a table that contains only one column, and I want you to find a python program that detects if there is an error in the column. 
                                The error are often spelling errors, where the unusual tokens are introduced in the phrase. Hence, please try to avoid exact match in the program so it can **generalize** to other data samples.

                                The main python function that's callable to solve the task is called "detect_error(input_string)" where "input_string" is the function parameter, which often is the column value. 
                                If you need to import the python packages, assume they are already installed in the target environment. 
                                Assuming the list of python packages are installed: ["SpellChecker", "regex", "nltk", "fractions", "sklearn", "Levenshtein", "math"].
                                If you need to use spellchecker, then please try to include domain-specific vocabularies.
                                Please find functions that can generalize to many other tables of the same task. Please avoid listing exact match to the words in the examples in your generated algorithm. 
                                Please return only the python function and nothing else. Include any imports. Please do not provide any comments. After generating the function string, please add escape symbol tab '\t' to format the code properly with indents.

                                I will provide you some input/output examples as references.
                                """},
                                {"role": "user", "content": f"""Instructions: {{instruction:}}, Examples: {{examples:}}"""}],

    "ERROR_DETECTION_SPELLING_INSTR": f"""I have a table with adult data, now I will provide rows of records with the column name: {{column_name}}, please write an appropriate python program that checks if there are any spelling errors. Return "yes" if there is spelling and "no" if there is none.""",
    "FUNCTION_REGENERATION": [{"role": "user", "content": f"""Instructions: \"The proposed function {{function_str}} if not working on the example. Can your provide another function string? Please return only the python function and nothing else. Include any imports. Do not provide any comments.\""""}],
    "REGEX_GENERATOR": [{"role": "system", "content": """You are an expert Regex user. I will give you a string transformation task. \
                            The task involves converting input strings to output strings. \
                            You will have to find the Regex pattern called "string_transformation_regex" that will generate the output string given the input string. \
                            Please return only the python function and nothing else. Include any imports. Do not provide any comments."""},\
                            {"role": "user", "content": f"""Instructions: {{instruction:}}"""}]
}

PYTHON_PROMPTS_ROUTE_DATA = {
    "TASK_PREFIX":[ {"role": "system", "content": "You are helpful assistant. Given the following input/output examples, generate a python function that converts input to output. Reason step-by-step about the what the function is supposed to do and then generate the function. You can assume that the following python libraries are available in the target environment."}],
    "TASK_RETRY":[ {"role": "user", "content": f"{{error_message}}"}],
    "STRING_TRANSFORMATION": [{"role": "user", "content": """You are an expert Python user. I will give you a string transformation task. \
                         The task involves converting input strings to output strings. You will need to generate two pieces of python code, one is the transformation function called "string_transformation", one is the validation function called "validate". \
                         
                         Assuming all of the following python packages are installed: ["regex", "fractions", "math", "pyproj", "BeautifulSoup", "geopy", "ummalqura", "mgrs", "pytz", "datetime", "calendar", "roman"],\
                          you can **import** them and please put all imports in the begining of the generated python program. You can also import other packages if the listed ones are not working, but try using listed packages first. \
                          When it comes to currency exchange, please do not use real-time currency, but figure out the exchange rate based on the examples I provide. \
                         
                         For the transformation, you will have to find the Python program called "string_transformation" that will generate the output string given the input string. \
                          Output can only be string and nothing else. \
                          Please try to understand the intentions of the input-output examples while generating the function. \
                          Please return only the python function and nothing else.  Include any imports. Do not provide any comments.\
                               
                         For the validation function, I want you to analyse the assumptions made for the inputs that are used by the transformation function, and generate a validation function in Python. \
                         The idea is that only inputs that can pass the validation function should be used for the transformation function. \
                         For example, when the instruction is 'converting mg to ml', then the input should contain numerical values and should not be negative. \
                         The validation function you generate should be also in python, likely using regex to match the input or if/else conditions to check heuristics. \
                         The validation function should return True when the input can go through the transformation function and return False if the transformation function shouldn't be applied to it. \
                         Please find the validation function and name it "validate". Only return True or False for the validation function. \
                         
                        
                          Here is a very simple example: \n
                          Input: "john.doe@example.com" \n 
                          Output: "Yes" \n 
                          Input: "jane_doe@example.com" \n 
                          Output: "Yes" 
                          Input: "example.com" \n
                          Output: "No" \n
                          string_transformation_code: \n
                          ```pythonimport re\ndef string_transformation(input_string):\n\tpattern = r'^[\w\.-]+@[\w\.-]+\.\w+$'\n\tif re.match(pattern, input_string):\n\t\treturn "Yes"\n\telse:
        return\n\t\t"No"```\n
                          validate_code: \n
                        ```pythonimport re\ndef validate(input_string):\n\tpattern = r'^[\w\.-]+@[\w\.-]+\.\w+$'\n\tif re.match(pattern, input_string):\n\t\treturn True\n\telse:
        return\n\t\tFalse"```\n
                        
                          If using regex, please come up with correct pattern and make sure the extraction will not be empty \
                          Now it's your turn. After generating the function string, please add escape symbol tab '\t' to format the code properly with indents like shown in the example. """},
                        {"role": "user", "content": f"""Instructions: {{instruction:}}, Examples: {{examples:}}"""}],

    "ENTITY_MATCHING": [{"role": "user", "content": """You are an expert in Python. \
                         I have an entity matching task to solve. I have pairs of rows and I want you to find a python program that checks if they match with each other. \
                         Please try to match using substring and regex preferably. \
                         You will need to generate two pieces of python code, one is the transformation function called "entity_matcher", one is the validation function called "validate". \
                         
                         \n
                        Assuming all of the following python packages are installed in the target environment: ["numpy", "regex", "fractions", "sklearn", "Levenshtein", "math", "pyproj", "BeautifulSoup", "geopy", "ummalqura", "mgrs", "pytz", "datetime", "calendar", "roman"], 
                        you can import them when it is necessary and please put all imports in the beginning of the generated python program. 
                        You can also import other packages if the listed ones are not working, but try using listed packages first. 

                         For the entity_matcher function:     
                        First consider using Jaccard, levenshtein, cosine similarity or any other edit distance function to solve the problem. 
                        Try to define a low threshold so that the generated function is **generalizable**.
                        Note that there can be missing values in some columns, and you can skip them for matching. All columns are connected by the symbol ^ in the input string.
                        Please do not train any machine learning classifier.
                        You can try to choose appropriate distance function with revevant columns based on the column meaning.
                        If you need to list some mappings, try to retrieve all entries first or write a function to retrieve it first then write the matching function. 
                        Try to generate functions that can generalize to many other data samples than the ones we provided here. 
                        \n
                        The main python function that's callable to solve the task is called "entity_matcher(input_string)" where "input_string" is the function parameter. 
                         
                        For the validate function:
                        I want you to analyse the assumptions made for the inputs that are used by the entity_matcher function, and generate a validation function in Python. \
                        The idea is that only inputs that can pass the validation function should be used for the entity_matcher function. \
                        For example, the proposed entity_matcher function might require the input to be certain format, or have certain values. \
                        The validation function you generate should be also in python, likely using regex to match the input or if/else conditions to check heuristics. \
                        The validation function should return True when the input can go through the transformation function and return False if the transformation function shouldn't be applied to it. \
                        Please find the validation function and name it "validate". Only return True or False for the validation function. \
                
                    One example entity matcher function string is: ```pythonimport re\ndef entity_matcher(input_string):\n\tpattern = r'^[\w\.-]+@[\w\.-]+\.\w+$'\n\tif re.match(pattern, input_string):\n\t\treturn "Yes"\n\telse: return\n\t\t"No"``` 
                    \n
                    One example validate function string is: ```pythonimport re\ndef validate(input_string):\n\tpattern = r'^[\w\.-]+@[\w\.-]+\.\w+$'\n\tif re.match(pattern, input_string):\n\t\treturn True\n\telse: return\n\t\tFalse"```\n .
                    Please return only the python function and nothing else. Include any imports. Please do **not** provide any comments. 
                    After generating the function string, please add escape symbol tab '\t' to format the code properly with indents like shown in the example."""},\
                        {"role": "user", "content": f"""Instructions: {{instruction:}}, Examples: {{examples:}}"""}],

    "DATA_IMPUTATION": [{"role": "user", "content": """You are a helpful code searcher. I have a data imputation task to solve. I have rows of records and I want you to find a python program that substituting missing values given the column name.

                    Please try to find python programs that uses logical conditions, substring and regex preferably, in combination of the column meanings. 
                         
                    You will need to generate two pieces of python code, one is the transformation function called "impute_data", one is the validation function called "validate". \

                    Assuming all of the following python packages are installed in the target environment: ["regex", "fractions", "sklearn", "Levenshtein", "math", "pyproj", "BeautifulSoup", "geopy", "ummalqura", "mgrs", "pytz", "datetime", "calendar", "roman"], you can import them when it is necessary and please put all imports in the beginning of the generated python program. 
                    You can also import other packages if the listed ones are not working, but try using listed packages first. 
                         
                    For the impute_data function:
                        - Try to define the range dynamically so that the generated program is **generalizable**. 
                        - You will likely need to extropolate the missing values based on the column meanings. So please do not only use exact match. If you have to, then specify it in the validation function so we can route the inputs.
                        - Please do not train any machine learning classifier.
                        - If you need to list some mappings, try to retrieve all entries first or write a function to retrieve it first then write the matching function. 
                        - Try to generate functions that can generalize to many other data samples than the ones we provided here.
                        - The main python function that's callable to solve the task is called "impute_data(input_string)" where "input_string" is the function parameter. 
                    
                    For the validate function:
                        - I want you to analyse the assumptions made for the inputs that are used by the impute_data function, and generate a validation function in Python. \
                        - If the impute_data function contains exact match or mappings, try to specify the conditions in the validation function. For example, only return true when the input is in the mapping.
                        - The idea is that only inputs that can pass the validation function should be used for the impute_data function. \
                        - For example, the proposed impute_data function might require the input to be certain format, or have certain values. \
                        - The validation function you generate should be also in python, likely using regex to match the input or if/else conditions to check heuristics. \
                        - The validation function should return True when the input can go through the transformation function and return False if the transformation function shouldn't be applied to it. \
                        - Please find the validation function and name it "validate". Only return True or False for the validation function. \
                     
                    One example function string is: pythonimport re\ndef impute_data(input_string):\n\tpattern = r'^[\w\.-]+@[\w\.-]+\.\w+$'\n\tif re.match(pattern, input_string):\n\t\treturn "Yes"\n\telse: return\n\t\t"No"
                    One example validate function string is: ```pythonimport re\ndef validate(input_string):\n\tpattern = r'^[\w\.-]+@[\w\.-]+\.\w+$'\n\tif re.match(pattern, input_string):\n\t\treturn True\n\telse: return\n\t\tFalse"```\n .
                         
                    Please return only the python function and nothing else. Include any imports. Please do **not** provide any comments. After generating the function string, please add escape symbol tab '\t' to format the code properly with indents like shown in the example.

                    """},
                    {"role": "user", "content": f"""Instructions: {{instruction:}}, Examples: {{examples:}}"""}],

    "ERROR_DETECTION_SPELLING": [{"role": "user", "content": """You are a helpful code generator. I have an error detection task to solve. I have rows of records from a table that contains only one column, and I want you to find a python program that detects if there is an error in the column. 
                                The error are often spelling errors, where the unusual tokens are introduced in the phrase. Hence, please try to avoid exact match in the program so it can **generalize** to other data samples.
                                You will need to generate two pieces of python code, one is the transformation function called "detect_error", one is the validation function called "validate". \
                                Assuming all of the following python packages are installed in the target environment: ["SpellChecker", "regex", "nltk", "fractions", "sklearn", "Levenshtein", "math"].
                                If you need to use spellchecker, then please try to include domain-specific vocabularies.
                                You can also import other packages if the listed ones are not working, but try using listed packages first. 

                                For the detect_error function:
                                    - The main python function that's callable to solve the task is called "detect_error(input_string)" where "input_string" is the function parameter, which often is the column value. 
                                    - Please find functions that can generalize to many other tables of the same task. Please avoid listing exact match to the words in the examples in your generated algorithm. 
                                    - Please return only the python function and nothing else. Include any imports. Please do not provide any comments. After generating the function string, please add escape symbol tab '\t' to format the code properly with indents.
                                
                                For the validate function:
                                    - I want you to analyse the assumptions made for the inputs that are used by the detect_error function, and generate a validation function in Python. \
                                    - The idea is that only inputs that can pass the validation function should be used for the detect_error function. \
                                    - For example, the proposed detect_error function might require the input to be certain format, or have certain values. \
                                    - The validation function you generate should be also in python, likely using regex to match the input or if/else conditions to check heuristics. \
                                    - The validation function should return True when the input can go through the transformation function and return False if the transformation function shouldn't be applied to it. \
                                    - Please find the validation function and name it "validate". Only return True or False for the validation function. \
                     
                                I will provide you some input/output examples as references.
                                Please return only the python function and nothing else. Include any imports. Please do **not** provide any comments. After generating the function string, please add escape symbol tab '\t' to format the code properly with indents like shown in the example.
                                """},
                                {"role": "user", "content": f"""Instructions: {{instruction:}}, Examples: {{examples:}}"""}],

    "ERROR_DETECTION_SPELLING_INSTR": f"""I have a table with adult data, now I will provide rows of records with the column name: {{column_name}}, please write an appropriate python program that checks if there are any spelling errors. Return "yes" if there is spelling and "no" if there is none.""",
}


SQL_PROMPTS_ROUTE_DATA = {
    "TASK_PREFIX": [ {"role": "system", "content": "You are helpful assistant. Given the following input/output examples, generate a DuckDB SQL macro that converts input to output. Reason step-by-step about the what the DuckDB SQL macro is supposed to do and then generate."}],
    "TASK_RETRY": [ {"role": "user", "content": f"{{error_message}}"}],
    "STRING_TRANSFORMATION": [{"role": "user", "content": """You are an DuckDB SQL expert. I will give you a string transformation task. \
                            The task involves converting input strings to output strings. \
                               
                            I want you to analyse the assumptions made for the inputs that are used by the transformation function, and generate a validation function in DuckDB sql macro.
                            The idea is that only inputs that can pass the validation function should be used for the transformation function.
                            For example, when the instruction is 'converting mg to ml', then the input should contain numerical values and should not be negative. 
                            The validation function you generate should be also in DuckDB sql macro, likely using regex to match the input or if/else conditions to check heuristics.
                            The function should return True when the input can go through the transformation function and return False if the transformation function shouldn't be applied to it.
                            
                            There are some additional information that you might need to generate the transformation and validation macros:
                            
                            - When it comes to currency exchange, please do not use real-time currency, but figure out the exchange rate based on the examples I provide.\
                            - You will have to use the CREATE MACRO statement from DuckDB, following is an example usage:\
                                -- create a macro with a common table expression \
                                -- (parameter names get priority over column names: disambiguate using the table name) \
                                CREATE or REPLACE MACRO string_transformation(a) AS (WITH cte AS (SELECT 1 AS a) SELECT cte.a + a FROM cte); \
                            - For the transformation function, please keep the name string_transformation fixed and only generate the expression in the brackets after CREATE OR REPLACE AS (). \
                            - For the validation function, please keep the name validate fixed and only generate the expression in the brackets after CREATE OR REPLACE AS (). \
                            - Please return **only** boolean values True or False from the validation function. The validation function should be able to cast **any** input to boolean. \
                            - If the input is a string, validate if the input string is properly formatted for the transformation function using regex, then check the correct range for the numerical part. \
                            - Make sure to only use DuckDB SQL to solve the task, no Python udfs or anything like that. \
                            - Please always consider the full format of the input to the transformation function. Such as '3 mg' instead of '3'.\
                            - You can for example use DuckDBs regex function as part of the program. \
                                Name	Description
                                regexp_extract_all(string, regex[, group = 0])	Split the string along the regex and extract all occurrences of group. \
                                regexp_extract(string, pattern , name_list);	If string contains the regexp pattern, returns the capturing groups as a struct with corresponding names from name_list. \
                                regexp_extract(string, pattern [, idx]);	If string contains the regexp pattern, returns the capturing group specified by optional parameter idx. \
                                regexp_full_match(string, regex)	Returns true if the entire string matches the regex. \
                                regexp_matches(string, pattern)	Returns true if string contains the regexp pattern, false otherwise. \
                                regexp_replace(string, pattern, replacement);	If string contains the regexp pattern, replaces the matching part with replacement. \
                                regexp_split_to_array(string, regex)	Alias of string_split_regex. Splits the string along the regex. \
                                regexp_split_to_table(string, regex)	Splits the string along the regex and returns a row for each part. \
                            - When use backslash symbol, please use only single backslash instead of two, so the query can be executed. \
                            - If using regex, please come up with correct pattern and make sure the extraction will not be empty \
                            
                            If you think the task can be solved using other DuckDB functions, please mention in the reasoning and use them correctly. \
                            The user will provide you input/output examples and sometimes instructions. """},
                            {"role": "user", "content": f"""Instructions: {{instruction}}, Examples: {{examples}}"""}]
}

# SQL_PROMPTS = {
#             "TASK_PREFIX": [ {"role": "system", "content": "You are helpful assistant. Given the following input/output examples, generate a DuckDB SQL macro that converts input to output. Reason step-by-step about the what the DuckDB SQL macro is supposed to do and then generate."}],
#             "TASK_RETRY": [ {"role": "user", "content": f"{{error_message}}"}],
#             "STRING_TRANSFORMATION": [{"role": "user", "content": """You are an DuckDB SQL expert. I will give you a string transformation task. \
#                                     The task involves converting input strings to output strings. \
#                                     When it comes to currency exchange, please do not use real-time currency, but figure out the exchange rate based on the examples I provide.\
#                                     You will have to use the CREATE MACRO statement from DuckDB, following is an example usage:\
#                                     -- create a macro with a common table expression \
#                                     -- (parameter names get priority over column names: disambiguate using the table name) \
#                                     CREATE or REPLACE MACRO string_transformation(a) AS (WITH cte AS (SELECT 1 AS a) SELECT cte.a + a FROM cte); \
#                                     Please keep the name string_transformation fixed and only generate the expression in the brackets after CREATE OR REPLACE AS (). \
#                                     Make sure to only use DuckDB SQL to solve the task, no Python udfs or anything like that. \
#                                     You can for example use DuckDBs regex function as part of the program. \
#                                     Name	Description
#                                     regexp_extract_all(string, regex[, group = 0])	Split the string along the regex and extract all occurrences of group. \
#                                     regexp_extract(string, pattern , name_list);	If string contains the regexp pattern, returns the capturing groups as a struct with corresponding names from name_list. \
#                                     regexp_extract(string, pattern [, idx]);	If string contains the regexp pattern, returns the capturing group specified by optional parameter idx. \
#                                     regexp_full_match(string, regex)	Returns true if the entire string matches the regex. \
#                                     regexp_matches(string, pattern)	Returns true if string contains the regexp pattern, false otherwise. \
#                                     regexp_replace(string, pattern, replacement);	If string contains the regexp pattern, replaces the matching part with replacement. \
#                                     regexp_split_to_array(string, regex)	Alias of string_split_regex. Splits the string along the regex. \
#                                     regexp_split_to_table(string, regex)	Splits the string along the regex and returns a row for each part. \
#                                     When use backslash symbol, please use only single backslash instead of two, so the query can be executed. \
#                                     If using regex, please come up with correct pattern and make sure the extraction will not be empty \
#                                     The user will provide you input/output examples and sometimes instructions. """},
#                                     {"role": "user", "content": f"""Instructions: {{instruction}}, Examples: {{examples}}"""}]

# }


SQL_PROMPTS = {
            "TASK_PREFIX": [ {"role": "system", "content": "You are helpful assistant. Given the following input/output examples, generate a DuckDB SQL macro that converts input to output. Reason step-by-step about the what the DuckDB SQL macro is supposed to do and then generate."}],
            "TASK_RETRY": [ {"role": "user", "content": f"{{error_message}}"}],
            "STRING_TRANSFORMATION": [{"role": "user", "content": """You are an DuckDB SQL expert. I will give you a string transformation task. \
                                    The task involves converting input strings to output strings. \
                                    When it comes to currency exchange, please do not use real-time currency, but figure out the exchange rate based on the examples I provide.\
                                    You will have to use the CREATE MACRO statement from DuckDB, following is an example usage:\
                                    -- create a macro with a common table expression \
                                    -- (parameter names get priority over column names: disambiguate using the table name) \
                                    CREATE or REPLACE MACRO string_transformation(a) AS (WITH cte AS (SELECT 1 AS a) SELECT cte.a + a FROM cte); \
                                    Please keep the name string_transformation fixed and only generate the expression in the brackets after CREATE OR REPLACE AS (). \
                                    Make sure to only use DuckDB SQL to solve the task, no Python udfs or anything like that. \
                                    You can for example use DuckDBs regex function as part of the program. \
                                    Name	Description
                                    regexp_extract_all(string, regex[, group = 0])	Split the string along the regex and extract all occurrences of group. \
                                    regexp_extract(string, pattern , name_list);	If string contains the regexp pattern, returns the capturing groups as a struct with corresponding names from name_list. \
                                    regexp_extract(string, pattern [, idx]);	If string contains the regexp pattern, returns the capturing group specified by optional parameter idx. \
                                    regexp_full_match(string, regex)	Returns true if the entire string matches the regex. \
                                    regexp_matches(string, pattern)	Returns true if string contains the regexp pattern, false otherwise. \
                                    regexp_replace(string, pattern, replacement);	If string contains the regexp pattern, replaces the matching part with replacement. \
                                    regexp_split_to_array(string, regex)	Alias of string_split_regex. Splits the string along the regex. \
                                    regexp_split_to_table(string, regex)	Splits the string along the regex and returns a row for each part. \
                                    When use backslash symbol, please use only single backslash instead of two, so the query can be executed. \
                                    If using regex, please come up with correct pattern and make sure the extraction will not be empty \
                                    Here are some DuckDB SQL syntax specifics you should be aware of:
                                    - DuckDB use double quotes (") for identifiers that contain spaces or special characters, or to force case-sensitivity and single quotes (') to define string literals
                                    - DuckDB can query CSV, Parquet, and JSON directly without loading them first, e.g. `SELECT * FROM 'data.csv';`
                                    - DuckDB supports CREATE TABLE AS (CTAS): `CREATE TABLE new_table AS SELECT * FROM old_table;`
                                    - DuckDB queries can start with FROM, and optionally omit SELECT *, e.g. `FROM my_table WHERE condition;` is equivalent to `SELECT * FROM my_table WHERE condition;`
                                    - DuckDB allows you to use SELECT without a FROM clause to generate a single row of results or to work with expressions directly, e.g. `SELECT 1 + 1 AS result;`
                                    - DuckDB supports attaching multiple databases, unsing the ATTACH statement: `ATTACH 'my_database.duckdb' AS mydb;`. Tables within attached databases can be accessed using the dot notation (.), e.g. `SELECT * FROM mydb.table_name syntax`. The default databases doesn't require the do notation to access tables. The default database can be changed with the USE statement, e.g. `USE my_db;`.
                                    - DuckDB is generally more lenient with implicit type conversions (e.g. `SELECT '42' + 1;` - Implicit cast, result is 43), but you can always be explicit using `::`, e.g. `SELECT '42'::INTEGER + 1;`
                                    - DuckDB can extract parts of strings and lists using [start:end] or [start:end:step] syntax. Indexes start at 1. String slicing: `SELECT 'DuckDB'[1:4];`. Array/List slicing: `SELECT [1, 2, 3, 4][1:3];`
                                    - DuckDB has a powerful way to select or transform multiple columns using patterns or functions. You can select columns matching a pattern: `SELECT COLUMNS('sales_.*') FROM sales_data;` or transform multiple columns with a function: `SELECT AVG(COLUMNS('sales_.*')) FROM sales_data;`
                                    - DuckDB an easy way to include/exclude or modify columns when selecting all: e.g. Exclude: `SELECT * EXCLUDE (sensitive_data) FROM users;` Replace: `SELECT * REPLACE (UPPER(name) AS name) FROM users;`
                                    - DuckDB has a shorthand for grouping/ordering by all non-aggregated/all columns. e.g `SELECT category, SUM(sales) FROM sales_data GROUP BY ALL;` and `SELECT * FROM my_table ORDER BY ALL;`
                                    - DuckDB can combine tables by matching column names, not just their positions using UNION BY NAME. E.g. `SELECT * FROM table1 UNION BY NAME SELECT * FROM table2;`
                                    - DuckDB has an inutitive syntax to create List/Struct/Map and Array types. Create complex types using intuitive syntax. List: `SELECT [1, 2, 3] AS my_list;`, Struct: `{{'a': 1, 'b': 'text'}} AS my_struct;`, Map: `MAP([1,2],['one','two']) as my_map;`. All types can also be nested into each other. Array types are fixed size, while list types have variable size. Compared to Structs, MAPs do not need to have the same keys present for each row, but keys can only be of type Integer or Varchar. Example: `CREATE TABLE example (my_list INTEGER[], my_struct STRUCT(a INTEGER, b TEXT), my_map MAP(INTEGER, VARCHAR),  my_array INTEGER[3], my_nested_struct STRUCT(a INTEGER, b Integer[3]));`
                                    - DuckDB has an inutive syntax to access struct fields using dot notation (.) or brackets ([]) with the field name. Maps fields can be accessed by brackets ([]).
                                    - DuckDB's way of converting between text and timestamps, and extract date parts. Current date as 'YYYY-MM-DD': `SELECT strftime(NOW(), '%Y-%m-%d');` String to timestamp: `SELECT strptime('2023-07-23', '%Y-%m-%d')::TIMESTAMP;`, Extract Year from date: `SELECT EXTRACT(YEAR FROM DATE '2023-07-23');`
                                    - Column Aliases in WHERE/GROUP BY/HAVING: You can use column aliases defined in the SELECT clause within the WHERE, GROUP BY, and HAVING clauses. E.g.: `SELECT a + b AS total FROM my_table WHERE total > 10 GROUP BY total HAVING total < 20;`
                                    - DuckDB allows generating lists using expressions similar to Python list comprehensions. E.g. `SELECT [x*2 FOR x IN [1, 2, 3]];` Returns [2, 4, 6].
                                    - DuckDB allows chaining multiple function calls together using the dot (.) operator. E.g.: `SELECT 'DuckDB'.replace('Duck', 'Goose').upper(); -- Returns 'GOOSEDB';`
                                    - DuckDB has a JSON data type. It supports selecting fields from the JSON with a JSON-Path expression using the arrow operator, -> (returns JSON) or ->> (returns text) with JSONPath expressions. For example: `SELECT data->'$.user.id' AS user_id, data->>'$.event_type' AS event_type FROM events;`
                                    - DuckDB has built-in functions for regex regexp_matches(column, regex), regexp_replace(column, regex), and regexp_extract(column, regex).
                                    - DuckDB has a way to quickly get a subset of your data with `SELECT * FROM large_table USING SAMPLE 10%;`
                                    I will provide you input/output examples and sometimes instructions. 
                                    Please return **only** the DuckDB sql macro code and nothing else. Put the DuckDB sql macro in the quoted block: 
                                       ```sql
                                        CREATE OR REPLACE MACRO string_transformation(a) AS (
                                             SELECT ...
                                        ```
                                       """},
                                    {"role": "user", "content": f"""Instructions: {{instruction}}, Examples: {{examples}}"""}],

            "ENTITY_MATCHING": [{"role": "user", "content": """You are a DuckDB SQL expert. I have an entity matching task to solve. I have pairs of rows and I want you to find a python program that checks if they match with each other. 
                    Please try to match using substring and regex preferably. 
                    First consider using Jaccard, levenshtein, cosine similarity or any other edit distance function to solve the problem. 
                    Try to define a low threshold so that the generated function is **generalizable**.
                    Note that there can be missing values in some columns, and you can skip them for matching. All columns are connected by the symbol ^ in the input string.
                    Please do not train any machine learning classifier.
                    You can try to choose appropriate distance function with revevant columns based on the column meaning.
                    
                    You will have to use the CREATE MACRO statement from DuckDB, following is an example usage:\
                    -- create a macro with a common table expression \
                    -- (parameter names get priority over column names: disambiguate using the table name) \
                    CREATE or REPLACE MACRO string_transformation(a) AS (WITH cte AS (SELECT 1 AS a) SELECT cte.a + a FROM cte); \
                    Please keep the name string_transformation fixed and only generate the expression in the brackets after CREATE OR REPLACE AS (). \
                    Make sure to only use DuckDB SQL to solve the task, no Python udfs or anything like that. \
                    You can for example use DuckDBs regex function as part of the program. \
                    Name	Description
                    regexp_extract_all(string, regex[, group = 0])	Split the string along the regex and extract all occurrences of group. \
                    regexp_extract(string, pattern , name_list);	If string contains the regexp pattern, returns the capturing groups as a struct with corresponding names from name_list. \
                    regexp_extract(string, pattern [, idx]);	If string contains the regexp pattern, returns the capturing group specified by optional parameter idx. \
                    regexp_full_match(string, regex)	Returns true if the entire string matches the regex. \
                    regexp_matches(string, pattern)	Returns true if string contains the regexp pattern, false otherwise. \
                    regexp_replace(string, pattern, replacement);	If string contains the regexp pattern, replaces the matching part with replacement. \
                    regexp_split_to_array(string, regex)	Alias of string_split_regex. Splits the string along the regex. \
                    regexp_split_to_table(string, regex)	Splits the string along the regex and returns a row for each part. \
                    When use backslash symbol, please use only single backslash instead of two, so the query can be executed. \
                    If using regex, please come up with correct pattern and make sure the extraction will not be empty \
                    Here are some DuckDB SQL syntax specifics you should be aware of:
                    - DuckDB use double quotes (") for identifiers that contain spaces or special characters, or to force case-sensitivity and single quotes (') to define string literals
                    - DuckDB can query CSV, Parquet, and JSON directly without loading them first, e.g. `SELECT * FROM 'data.csv';`
                    - DuckDB supports CREATE TABLE AS (CTAS): `CREATE TABLE new_table AS SELECT * FROM old_table;`
                    - DuckDB queries can start with FROM, and optionally omit SELECT *, e.g. `FROM my_table WHERE condition;` is equivalent to `SELECT * FROM my_table WHERE condition;`
                    - DuckDB allows you to use SELECT without a FROM clause to generate a single row of results or to work with expressions directly, e.g. `SELECT 1 + 1 AS result;`
                    - DuckDB supports attaching multiple databases, unsing the ATTACH statement: `ATTACH 'my_database.duckdb' AS mydb;`. Tables within attached databases can be accessed using the dot notation (.), e.g. `SELECT * FROM mydb.table_name syntax`. The default databases doesn't require the do notation to access tables. The default database can be changed with the USE statement, e.g. `USE my_db;`.
                    - DuckDB is generally more lenient with implicit type conversions (e.g. `SELECT '42' + 1;` - Implicit cast, result is 43), but you can always be explicit using `::`, e.g. `SELECT '42'::INTEGER + 1;`
                    - DuckDB can extract parts of strings and lists using [start:end] or [start:end:step] syntax. Indexes start at 1. String slicing: `SELECT 'DuckDB'[1:4];`. Array/List slicing: `SELECT [1, 2, 3, 4][1:3];`
                    - DuckDB has a powerful way to select or transform multiple columns using patterns or functions. You can select columns matching a pattern: `SELECT COLUMNS('sales_.*') FROM sales_data;` or transform multiple columns with a function: `SELECT AVG(COLUMNS('sales_.*')) FROM sales_data;`
                    - DuckDB an easy way to include/exclude or modify columns when selecting all: e.g. Exclude: `SELECT * EXCLUDE (sensitive_data) FROM users;` Replace: `SELECT * REPLACE (UPPER(name) AS name) FROM users;`
                    - DuckDB has a shorthand for grouping/ordering by all non-aggregated/all columns. e.g `SELECT category, SUM(sales) FROM sales_data GROUP BY ALL;` and `SELECT * FROM my_table ORDER BY ALL;`
                    - DuckDB can combine tables by matching column names, not just their positions using UNION BY NAME. E.g. `SELECT * FROM table1 UNION BY NAME SELECT * FROM table2;`
                    - DuckDB has an inutitive syntax to create List/Struct/Map and Array types. Create complex types using intuitive syntax. List: `SELECT [1, 2, 3] AS my_list;`, Struct: `{{'a': 1, 'b': 'text'}} AS my_struct;`, Map: `MAP([1,2],['one','two']) as my_map;`. All types can also be nested into each other. Array types are fixed size, while list types have variable size. Compared to Structs, MAPs do not need to have the same keys present for each row, but keys can only be of type Integer or Varchar. Example: `CREATE TABLE example (my_list INTEGER[], my_struct STRUCT(a INTEGER, b TEXT), my_map MAP(INTEGER, VARCHAR),  my_array INTEGER[3], my_nested_struct STRUCT(a INTEGER, b Integer[3]));`
                    - DuckDB has an inutive syntax to access struct fields using dot notation (.) or brackets ([]) with the field name. Maps fields can be accessed by brackets ([]).
                    - DuckDB's way of converting between text and timestamps, and extract date parts. Current date as 'YYYY-MM-DD': `SELECT strftime(NOW(), '%Y-%m-%d');` String to timestamp: `SELECT strptime('2023-07-23', '%Y-%m-%d')::TIMESTAMP;`, Extract Year from date: `SELECT EXTRACT(YEAR FROM DATE '2023-07-23');`
                    - Column Aliases in WHERE/GROUP BY/HAVING: You can use column aliases defined in the SELECT clause within the WHERE, GROUP BY, and HAVING clauses. E.g.: `SELECT a + b AS total FROM my_table WHERE total > 10 GROUP BY total HAVING total < 20;`
                    - DuckDB allows generating lists using expressions similar to Python list comprehensions. E.g. `SELECT [x*2 FOR x IN [1, 2, 3]];` Returns [2, 4, 6].
                    - DuckDB allows chaining multiple function calls together using the dot (.) operator. E.g.: `SELECT 'DuckDB'.replace('Duck', 'Goose').upper(); -- Returns 'GOOSEDB';`
                    - DuckDB has a JSON data type. It supports selecting fields from the JSON with a JSON-Path expression using the arrow operator, -> (returns JSON) or ->> (returns text) with JSONPath expressions. For example: `SELECT data->'$.user.id' AS user_id, data->>'$.event_type' AS event_type FROM events;`
                    - DuckDB has built-in functions for regex regexp_matches(column, regex), regexp_replace(column, regex), and regexp_extract(column, regex).
                    - DuckDB has a way to quickly get a subset of your data with `SELECT * FROM large_table USING SAMPLE 10%;`
                    I will provide you input/output examples and sometimes instructions. 
                    Please return **only** the DuckDB sql macro code and nothing else. Put the DuckDB sql macro in the quoted block: 
                    ```sql
                    CREATE OR REPLACE MACRO input_output_converter(a) AS (
                            SELECT ...
                    ```
                    Please return only the DuckDB sql macro and nothing else. Include any imports. Please do **not** provide any comments. 
                    After generating the function string, please add escape symbol tab '\t' to format the code properly with indents like shown in the example."""},\
                        {"role": "user", "content": f"""Instructions: {{instruction:}}, Examples: {{examples:}}"""}]
}

def get_prompts_set(lang, use_data_router=True):
    if lang == "python" and use_data_router:
        return PYTHON_PROMPTS_ROUTE_DATA
    elif lang == "python" and not use_data_router:
        return PYTHON_PROMPTS
    elif lang == "sql" and use_data_router:
        return SQL_PROMPTS_ROUTE_DATA
    elif lang == "sql" and not use_data_router:
        return SQL_PROMPTS
    else:
        raise NotImplementedError

class PromptTemplate:
    def __init__(self, prompt) -> None:
        self.prompt = prompt

    def execute(self):
        pass

    def reset(self):
        self.inital_prompt = ""
        self.message_history = []

    def assign_new_prompt(self, new_prompt: List) -> None:
        self.prompt = new_prompt

    def add_prompt_incremental(self, prompt_to_add: Dict) -> None:
        self.prompt.append(prompt_to_add) 








from ..code_generator.python_generator import PythonGenerator
from ..code_generator.sql_generator import SQLGenerator
import numpy as np

def generate_code_logic(lang:str, task:str, instruction:str, 
                        samples:list, api_key:str, hyperparams:dict):
    functions_pool = []
    router_functions_pool = []
    perfs = []
    use_data_router = hyperparams["use_data_router"]
    num_trials = hyperparams["num_trials"]
    num_retry = hyperparams["num_retry"]
    seed = hyperparams["seed"]
    num_iter = hyperparams["num_iterations"]
    llm = hyperparams["llm"]

    if lang == "python":
        generator = PythonGenerator(task=task, llm=llm, use_data_router=use_data_router, api_key=api_key)
        if task == "data_transformation":
            dummy_func = "def string_transformation(input_string): return None"
        elif task == "entity_matching":
            dummy_func = "def entity_matcher(input_string): return None"
        elif task == "data_imputation":
            dummy_func = "def impute_data(input_string): return None"
        elif task == "error_detection_spelling":
            dummy_func = "def detect_error(input_string): return None"
        else:
            raise NotImplementedError
    elif lang == "sql":
        generator = SQLGenerator(task=task, llm=llm, use_data_router=use_data_router, api_key=api_key)
        dummy_func = "SELECT 'None'"
    else:
        raise NotImplementedError
    # We do N trials 
    for trial_num in range(num_trials):
        seed = seed + trial_num
        np.random.seed(seed)
        num_try = 0
        acc = None
        t_func = "No function"
        v_func = None
        while "No function" in t_func:
            if num_try >= num_retry:
                # use dummy function string
                t_func = dummy_func
                continue
            # samples should be a list of dictionaries that shows {""input": "sample1", "output": "label1"}
            demonstrations = samples
            # demonstrations, demon_df = sample_demonstrations(args, task, train_data)
            result = generator.pipeline(instruction=instruction, examples=demonstrations, depth=num_iter)
            t_func = result["transformation_code"]
            v_func = result["validation_code"]
            acc = result["acc"]
            num_try += 1
            seed += 1234
            np.random.seed(seed)
        if use_data_router:
            router_functions_pool.append(v_func)
        functions_pool.append(t_func)
        perfs.append(acc)
        # logger.info("Learned function for trial {} is {}".format(trial_num, t_func))
        # if use_data_router:
            # logger.info("Validation function is {}".format(v_func))
        
    # if there are multiple functions, we can use the one with the highest accuracy
    if len(functions_pool) > 1:
        best_idx = np.argmax(perfs)
        best_func = functions_pool[best_idx]
        if use_data_router:
            best_router_func = router_functions_pool[best_idx]
        else:
            best_router_func = None
    else:
        best_func = functions_pool[0]
        if use_data_router:
            best_router_func = router_functions_pool[0]
        else:
            best_router_func = None
    # logger.info("Best function is {}".format(best_func))
    # logger.info("Best router function is {}".format(best_router_func))
    return best_func, best_router_func


from typing import List
from ..llm_pr.llmpr import LLMPR


def label_data_logic(instruction: str, task:str, demonstrations: List[dict],
                      samples_to_label: List[dict], api_key: str, hyperparams: dict) -> List[str]:
    """
    Generate labels for the input samples based on the provided instruction.
    """
    print("Labeling data logic...")
    print(f"Instruction: {instruction}")
    print(f"demonstration: {demonstrations}")

    llm = hyperparams["llm"]
    
    # Here we would typically call an LLM to generate labels based on the instruction and input samples
    # use llmpr to generate labels
    llm_pr = LLMPR(task=task, api_key=api_key, model_name=llm)
    transformed_samples = llm_pr.pipeline_no_eval(instruction, demonstrations, samples_to_label)
    return transformed_samples